function [xopt, fopt, stat] = ro_cmaes_kriging_additional_stats(fitnessfct, N, lb, ub, delta, stopeval, epsilon, param)
% [xopt, fopt, stat] = ro_cmaes_kriging_additional_stats(fitnessfct, N, lb, ub, delta, stopeval, epsilon, param)
%
%   Implementation of a CMA-ES for robust optimization using
%   Kriging metamodeling. This implementation also administrates
%   additional statistics for algorithm performance assessment.
%
% Input:
%   fitnessfct     - a handle to the fitness function
%   N              - the number of dimensions
%   lb             - the lower bounds of the box constraints
%                    this is a vector of length N
%   ub             - the upper bounds of the box constraints
%                    this is a vector of length N
%   delta          - this is a vector of length N containing
%                    information about the noise-distribution
%                    of the noise in the desigb variables
%   optimizer      - a function handle to the optimization method
%                    used for this robust optimization
%   stopeval       - the number of function evaluations used
%                    by the optimization algorithm
%   epsilon        - if applicable: the epsilon determines
%                    the desired accuracy in the search space
%                    for approximating the optimum (epsilon
%                    should be greater than or equal to 0)
%                    if not applicable: then set it to -1
%   param          - a MATLAB struct containing algorithm
%                    specific input parameters
%
% Output:
%   xopt           - a vector containing the location of the optimum
%   fopt           - the objective function value of the optimum
%   stat           - a MATLAB struct containing algorithm
%                    specific output statistics
%
% Author: Johannes W. Kruisselbrink
% Last modified: November 27, 2009

	% Test epsilon setting
	if (epsilon ~= -1)
		warning('epsilon setting ignored')
	end

	% Do plotting
	if (isfield(param, 'doplot'))
		doplot = param.doplot;
	else
		doplot = true;
	end

	% Box Constraint Handling function
	if (isfield(param, 'bch'))
		bch = param.bch;
	else
		bch = @(x, lb, ub) x;
	end

	% Specifies the number of samples per evaluation (default: m=10)
	if (isfield(param, 'm'))
		m = param.m;
	else
		m = 10;
	end

	% Specifies the number samples for building the Kriging model
	if (isfield(param, 'kriging_method'))
		kriging_method = param.kriging_method;
	else
		kriging_method = 'normal';
	end

	% Specifies the number samples for building the Kriging model
	if (isfield(param, 'n_krig'))
		n_krig = param.n_krig;
	else
		n_krig = max(6, 2 * N);
	end

	warning off

	% Initialize
	xmean = lb + (ub - lb) .* rand(N,1);
	sigma = (mean(ub - lb)) / sqrt(N);

	% Set lambda, mu and the weights for recombination
	lambda = 4 + floor(3 * log(N));
	mu = floor(lambda / 2);
	weights = log(mu + 1) - log(1:mu)';
	weights = weights / sum(weights);
	mueff = sum(weights)^2 / sum(weights.^2);

	% Set parameters
	cc = 4 / (N + 4);
	cs = (mueff + 2) / (N + mueff + 3);
	mucov = mueff;
	ccov = (1 / mucov) * 2 / (N + 1.4)^2 + (1 - 1/mucov) * ((2 * mueff - 1) / ((N + 2)^2 + 2 * mueff));
	damps = 1 + 2 * max(0, sqrt((mueff - 1) / (N + 1)) - 1) + cs;

	% Initialize matrices and vectors of the CMA-ES
	pc = zeros(N,1);
	ps = zeros(N,1);
	B = eye(N,N);
	D = eye(N,N);
	C = B * D * (B * D)';
	chiN = N^0.5 * (1 - 1 / (4 * N) + 1 / (21 * N^2));

	% Initialize archive
	A = zeros(N, stopeval);
	Af = zeros(1, stopeval);
	Asize = 0;

	% Statistics
	evalcount = 0;
	xopt = xmean;
	fopt = Inf;
	stat.histx = zeros(N, stopeval);
	stat.histf = zeros(1, stopeval);
	stat.histsigma = zeros(1, stopeval);
	stat.histC = zeros(N, N, stopeval);
	stat.histdiagD = zeros(N, stopeval);
	stat.histstds = zeros(N, stopeval);
	stat.evalcount = 0;
	stat.gencount = 0;
	stat.name = ['RO CMA-ES Kriging (', kriging_method, ')'];

	stat.model_construction_time = [];
	stat.archive_selection_time = [];
	stat.num_extra_samples = [];

	while evalcount < stopeval

		% Statistics administration
		mark_evalcount = evalcount;

		% Generate and evaluate lambda offspring
		for k=1:lambda,

			zo(:,k) = randn(N,1);
			xo(:,k) = xmean + sigma * (B * D * zo(:,k));
			xo(:,k) = feval(bch, xo(:,k), lb, ub);

			% Tricky: for cumulative stepsize adaptation you
			% probably want to recompute zo(:,k) when xo(:,k)
			% is changed by the box constraint handling method
			zo(:,k) = pinv(D) * pinv(B) * ((xo(:,k) - xmean) / sigma);

			% Statistics administration
			tic

			% Select from archive
			[xs, ys, xextra] = feval(@lhsu_reference_nb_select, xo(:,k), A(:,1:Asize), Af(1:Asize), delta, n_krig);

			% Statistics administration
			stat.archive_selection_time(end + 1) = toc;

			% Evaluate the extra sampling points
			extra_evaluations = size(xextra, 2);
			yextra = zeros(1, extra_evaluations);
			for i = 1 : extra_evaluations
				yextra(i) = feval(fitnessfct, xextra(:,i)');

				% Statistics administration
				evalcount = evalcount + 1;
				stat.histx(:,evalcount) = xopt;
				stat.histf(evalcount) = fopt;
				stat.histsigma(evalcount) = sigma;
				stat.histC(:,:,evalcount) = C;
				stat.histdiagD(:,evalcount) = diag(D);
				stat.histstds(:,evalcount) = sigma * sqrt(diag(C));
				stat.evalcount = evalcount;
			end
			xs = [xs, xextra];
			ys = [ys, yextra];

			% Statistics administration
			stat.num_extra_samples(end + 1) = extra_evaluations;

			% Update archive
			A = [xextra, A];
			Af = [yextra, Af];
			Asize = Asize + length(extra_evaluations);

			% Normalize points for Kriging, map to an interval between 0 and 100
			xmin = min(xs');
			xmax = max(xs');
			scalek = 100 ./ (xmax - xmin);
			xk = (xs' - repmat(xmin, size(xs,2),1)) .* repmat(scalek, size(xs,2), 1);
			yk = (ys' - min(ys)) * (100 ./ (max(ys) - min(ys)));

			% Statistics administration
			tic

			if (strcmp(kriging_method, 'normal'))
				% Construct simple Kriging function f_krig
				[beta_k, s_k, theta_k, C_k, invC_k, lambda_k] = calibrate_kriging(xk, yk);
				f_hat = @(X) kriging_estimate(xk, beta_k, s_k, theta_k, C_k, invC_k, lambda_k, (X - xmin) .* scalek) / (100 / (max(ys) - min(ys))) + min(ys);
			else
				% Obtain the correlation matrix from the covariance matrix
				A_mk = pinv(C) * pinv(C)';

				% Apply Kriging with the Mahalanobis-kernel function
				[beta_mk, s_mk, theta_mk, C_mk, invC_mk, lambda_mk] = calibrate_mahal_kriging(xk, yk, A_mk);
				f_hat = @(X) mahal_kriging_estimate(xk, beta_mk, s_mk, theta_mk, C_mk, invC_mk, lambda_mk, (X - xmin) .* scalek, A_mk) / (100 / (max(ys) - min(ys))) + min(ys);
			end

			% Statistics administration
			stat.model_construction_time(end + 1) = toc;

			% Calculate robustness approximation on metamodel
			xmci = lhsu(m, N, (xo(:,k) - delta), (xo(:,k) + delta))';
			feff = zeros(1,m);
			for i = 1 : m
				feff(i) = f_hat(xmci(:,i)');

				% Statistics administration
				freal(i) =  feval(fitnessfct, xmci(:,i)');
			end

			% Compute the fitness
			fo(k) = mean(feff);

			% Statistics administration
			foreal(k) = mean(freal);

		end

		% Sort by fitness and compute weighted means into xmean and zmean
		[sortf, sortindex] = sort(fo); % M I N I M I Z A T I O N
		xmean = xo(:,sortindex(1:mu)) * weights;
		zmean = zo(:,sortindex(1:mu)) * weights;

		% Cumulation: Update evolution paths
		ps = (1 - cs) * ps + sqrt(cs * (2 - cs) * mueff) * (B * zmean);
		hsig = norm(ps) / sqrt(1 - (1 - cs)^(2 * evalcount / lambda)) / chiN < 1.4 + 2/(N + 1);
		pc = (1 - cc) * pc + hsig * sqrt(cc * (2 - cc) * mueff) * (B * D * zmean);

		% Adapt covariance matrix C
		C = (1 - ccov) * C + ccov * (1 / mucov) * (pc * pc' + (1-hsig) * cc * (2 - cc) * C) + ccov * (1-1/mucov) * (B * D * zo(:,sortindex(1:mu))) * diag(weights) * (B * D * zo(:,sortindex(1:mu)))';

		% Adapt step size sigma
		sigma = sigma * exp((cs / damps) * (norm(ps) / chiN - 1));

		% Update B and D from C
		C = triu(C) + triu(C,1)';
		[B,D] = eig(C);
		D = diag(sqrt(diag(D)));

		% Statistics administration
		stat.gencount = stat.gencount + 1;
		stat.approximation_error(stat.gencount) = sum(abs(fo - foreal));
		stat.evalvsgen(stat.gencount) = evalcount - mark_evalcount;
		xopt = xo(:, sortindex(1));
		fopt = fo(sortindex(1));

		% Plot statistics
		if (doplot)
			clf
			plot_run_statistics(stat)
			drawnow()
		end

	end

	warning on
